@@ -1,8 +1,9 @@
 package Bitcoin::Crypto::Exception;
 
-our $VERSION = "1.001";
+our $VERSION = "1.002";
 
 use v5.10;
+use strict;
 use warnings;
 use Moo;
 use Types::Standard qw(Maybe Str ArrayRef);
@@ -152,6 +153,12 @@ sub as_string
 	use parent -norequire, "Bitcoin::Crypto::Exception";
 }
 
+{
+
+	package Bitcoin::Crypto::Exception::Bech32Type;
+	use parent -norequire, "Bitcoin::Crypto::Exception";
+}
+
 {
 
 	package Bitcoin::Crypto::Exception::Bech32InputChecksum;
